  require 'stringio'
  require 'fileutils'

# ---

module Raw
  def UInt(lengthInBytes)
    uint = 0
    to_s.slice(0, lengthInBytes).reverse.each_byte { |byte| uint = uint << 8 | byte }
    uint
  end

    def DWord
      UInt 4
    end

  def RawUInt(lengthInBytes)
    intVal = to_i

    [].fill(0, lengthInBytes) { |byte|
      bits = (byte * 8)
      ( (intVal & (0xFF << bits)) >> bits ).chr
    }.join
  end

    def RawDWord
      RawUInt 4
    end
end

  class String
    include Raw

    SlashedChars = {"\0" => '\0', "\t" => '\t', "\r" => '\r', "\n" => '\n'}

    def DWord!
      sub!( /^.*$/, DWord().to_s )
    end

    def AddSlashes
      gsub('\\', '\\\\\\\\').gsub( /([\x00-\x1F])/ ) { |match|
        SlashedChars.include?(match) ? SlashedChars[match] : ("\\x%.2X" % match.ord)
      }
    end

      def Unslash
          slashedToChar = SlashedChars.invert
          chars = slashedToChar.collect { |slashed, raw| Regexp.escape(slashed) }.join('|')

        gsub( /(\\*)(#{chars})/i ) { |fullMatch|
          $1.length % 2 == 0 ? slashedToChar[$2] : fullMatch
        }
        .gsub( /(\\*)\\x([0-9a-f]{2})/i ) { |fullMatch|
          $1.length % 2 == 0 ? $2.to_i(16).chr : fullMatch
        }
        .gsub('\\\\', '\\')
      end
  end

  class Fixnum
    include Raw
  end

class IO
  def ReadDWordArray(count)
    ReadUIntArray(count, 4)
  end

    def ReadUIntArray(count, uintLength)
      [].fill(0, count) { read(uintLength).UInt(uintLength) }
    end

  def WriteDWordArray(array)
    WriteUIntArray array, 4
  end

    def WriteUIntArray(array, uintLength)
      array.each { |item| WriteUInt item, uintLength }
    end

  def WriteDWord(dword)
    WriteUInt dword, 4
  end

    def WriteUInt(uint, uintLength)
      write uint.RawUInt uintLength
    end
end

def HexDumpOf(obj)
  obj.collect { |item| hexes << '%.2X' % item.to_i }.join(' ')
end


# --- MES-scripts related code ---

class Mes
  FileSignature = "\x66\x37"
  MaxLineLength = 54
  # Hardcoded in the game engine, longer causes a crash (1 byte reserved for
  # NUL, but since it's encoded in pairs of bytes in ShiftJIS - make it 2).
  StrLenWarningThreshold = 0x400-2

  attr_accessor :offsetCount, :offsets, :signature, :headerSize, :bytecode

  def self.LoadFromStream(script)
    mes = self.new
    mes.LoadFromStream(script)
    mes
  end

  def LoadFromStream(script)
    @offsetCount = script.read(4).DWord
    # There's no way to get size of stream opened by File.open? O_o
    if @offsetCount * 4 > File.size(script.path)
      throw 'Invalid script, too many offsets (%d)' % @offsetCount
    end

    @offsets = script.ReadDWordArray(@offsetCount)

    @signature = script.read(2)
    if @signature != FileSignature
      throw 'Invalid script signature "%s"' % signature
    end

      #   Those 2 bytes in FileSignature are actually part of bytecode
      # (at least offset table has offsets relative to position before them).
      script.seek -2, IO::SEEK_CUR

    @headerSize = script.pos
    @bytecode = MesBytecode.open(script.read, 'r+b')
  end

    def SaveToStream(script)
      script.truncate 0

      script.WriteDWord @offsetCount
      script.WriteDWordArray @offsets

      @bytecode.rewind
      IO.copy_stream @bytecode, script
    end

  def StripTexts
    text = []
    prevCmd = nil

    while !@bytecode.eof?
      prefix = ''
      str = nil
      strFollows = false

        cmdPos = headerSize + @bytecode.pos
        byte = @bytecode.readbyte

        case byte
          when 0x4E then prefix = '.'; strFollows = 'crypted'
          when 0x4D then prefix = '<'; strFollows = 'crypted'
          when 0x4F then prefix = '  >>  '; strFollows = 'crypted'
          when 0x34 then prefix = '@'; strFollows = true
          when 0x47 then prefix = 'if the following selected, go to @'; strFollows = true
          when 0x38 then prefix = 'exec '; strFollows = true
          when 0x30 then prefix = 'bkg '; strFollows = true
          when 0x35 then prefix = 'sprite '; strFollows = true
          when 0x31,
               0x48 then prefix = 'load CRX '; strFollows = true
          #   0x07 and 0x01 are two different commands but for partial
          # decompilation's sake this combo will give more accurate results.
          when 0x01 then prefix = 'fade-in & clrscr' if prevCmd == 0x07
        end

      case strFollows
      when true then      str = @bytecode.ReadString
      when 'crypted' then str = @bytecode.ReadStringAndDecrypt
      end

      if !prefix.to_s.empty? and (strFollows == false or !str.to_s.empty?)
        text << "%.3X | %s%s" % [cmdPos, prefix, str.to_s.AddSlashes]
      end

        prevCmd = byte
    end

    text
  end

  def UpdateUsing(texts)
    linesUpdated = 0

      globalDelta = 0

      texts.each("\n") { |line|
        if line.to_s.strip =~ /^([0-9a-f]+)\s*\|\s*(?:\.|<|>>)\s*(.+)$/i
          cmdOffset, str = $1.to_i(16), $2

            cmdOffset += globalDelta

            # Line format is either "original_text=translation" or simply "translation".
            # Only first = is removed.
            str.sub! /^.*?=/, ''
            str = str.to_s.strip.Unslash

          if str != '' && cmdOffset != 0
            @bytecode.seek cmdOffset - headerSize
            cmd = @bytecode.readbyte

            if [0x4D, 0x4E, 0x4F].include? cmd
              str = PrepareStrForMES str, cmd == 0x4E
              strOffset = @bytecode.pos

              oldStr = @bytecode.ReadStringAndDecrypt
              delta = str.length - oldStr.length

                if delta != 0
                  @bytecode.seek strOffset + oldStr.length
                    followingBytes = @bytecode.read
                  @bytecode.seek strOffset + str.length
                    @bytecode.write followingBytes

                  @bytecode.truncate @bytecode.size + delta if delta < 0
                end

              @bytecode.seek strOffset
              @bytecode.CryptAndWriteString str

                @offsets.collect! { |labelOffset|
                  # "- 1" for command opcode length (1 byte).
                  labelOffset += delta if labelOffset > strOffset - 1
                  labelOffset
                }

              globalDelta += delta
              linesUpdated += 1
            else
              print "Skipped non-string instruction at %.3X\n" % cmdOffset
            end
          end
        end
      }

    linesUpdated
  end

    def PrepareStrForMES(str, protectASCII)
      # Remove meme markers.
      str.gsub! /\$/, ''

      # The game engine doesn't handle wrapping, expecting the lines to be
      # pre-wrapped. However, even though the hook does support variable-width
      # fonts - we can't pre-wrap lines here in Ruby without mindboggling code
      # (because each line's length depends on GetTextExtentPoint32)
      # so the hook is configured to use a fixed-width font and here we just
      # wrap at a specific column (much like the game engine originally does).
      str = WordWrap str, MaxLineLength

      # Replace Japanese ellipsis with dots for better look when multiple ellipses are adjacent.
      #str.gsub!(/\.\.\.+|(\x81\x63)+/) { |s| s[0] == '.' ? "\x81\x63" : '...' }
      str.gsub! /(\x81\x63)+/, '...'

        passthruOnce = false
        sizeUsed = addSize = false

      str = str.split('').collect { |char|
        if passthruOnce
          addSize = char if addSize
          passthruOnce = false
          char
        else
          if char == ' '
            [0x81.chr, 0x40.chr]
          elsif char.ord < 0x20
            char  # e.g. \n
          elsif char.ord < 0x81 && char != '@'
            # The game engine isn't always (depending on opcode) friendly for
            # non-SJS chars (which are not 2 byte long) and skips the lower byte.
            #
            # The '@' char is special: it changes the font size and must be followed
            # by one of: s n m l - both '@' and that char must be 1-byte, else,
            # or when no supported char follows, '@' is silently ignored.
            protectASCII ? [char, '_'] : char
          else
            addSize = char == '@'
            sizeUsed = true if addSize
            passthruOnce = true
            char
          end
        end
      }.flatten.join

      # '@' affects all the following text including subsequent screens
      # which causes visual glitches so we always close off with @n.
      str += '@n' if sizeUsed

        if str.length > StrLenWarningThreshold
          # str snippet will show up as junk in the console. Redirect output to
          # a text file and open it as Shift-JIS (e.g. using Notepad 2e).
          print "  String is too long (%d bytes), truncated down to %d. %s\n" %
                [str.length, StrLenWarningThreshold, str[0, 40]]
          str = str[0, StrLenWarningThreshold - 1].strip
        end

      str
    end

      def WordWrap(str, lineLength)
        # Game texts are originally aligned so for quoted texts using any of
        # 「（『 all lines except the first start with a full-width space (8140);
        # for non-quoted texts only the first line starts with a space.
        # Just like with wrapping this is not handled by the engine on runtime.
        # We skip possible '@' size specifier as it's invisible.
        indent = str.match(/^(@.)?\x81[\x69\x75\x77]/) ? "\x81\x40" : ''
        str = "\x81\x40" + str if indent.empty?
        # str is 8bit (Shift-JIS), not UTF-8, so can't use /.../u and instead
        # treat all 7bit symbols as 1 byte and all others as 2 byte (MBCS).
        # This means lineLength is a number of symbols thus working for both
        # English and non-English texts even though str has varied number of
        # bytes per character. The method is not ideal as it produces English
        # lines with 2 chars above lineLength and non-English - with 3 chars.
        "#{str} ".gsub(/(([!-~]|..)([ -~]|..){1,#{lineLength-1}})(\s+)/, "\\1\n#{indent}").gsub(/^\s*|(\x81\x40|\s)*$/, '')
      end
end

  class MesBytecode < StringIO
    MaxStrLength = 2000

    def ReadString
      readline("\0").rstrip
    end

    def ReadStringAndDecrypt
      DecryptString( ReadString() )
    end

      def DecryptString(encodedStr)
        return nil if encodedStr.length > MaxStrLength

        encodedStr.each_byte.collect { |code|
          # 0xE0 = (0xFF + 1) (<= first impossible value) - 0x20
          (code >= 0xE0 ? code - 0xE0 : code + 0x20).chr
        }.join
      end

    def WriteString(str)
      write(str + "\0")
    end

    def CryptAndWriteString(str)
      WriteString( CryptString(str) )
    end

      def CryptString(str)
        return nil if str.length > MaxStrLength

        str.each_byte.collect { |code|
          case code
          when 0x00..0x1F then code + 0xE0
          # 0x20 - 0x20 = 0x00 (terminator for Shift-JIS-Z)
          when 0x20 then '_'.ord - 0x20
          else code - 0x20
          end.chr
        }.join
      end
  end

  def StripFromFile(scriptFile)
    File.open(scriptFile, 'rb') { |script| StripFromStream script }
  end

    def StripFromStream(script)
      Mes.LoadFromStream(script).StripTexts
    end

def StripAndSaveFromFiles(files)
    filesWithText = 0
    totalLines = 0
    allLines = []

  files.each { |file|
    print "Stripping texts from %s...\n" % File.basename(file)

    text = StripFromFile file
    allLines += text

    unless text.empty?
      print "  Found %d lines.\n" % text.count
      filesWithText += 1
      totalLines += text.count

      textFile = File.dirname(file) + '/' + File.basename(file, '.mes') + '.sjs'
      File.open(textFile, 'wb') { |f| f.write( text.join("\r\n") ) }

        FileUtils.touch file  # necessary for changed file detection in UpdateScripts()
    end
  }

    uniqueLines = allLines.collect { |line| line.sub(/^.+\|/, '') }.uniq.length

  print "\nDone, %d lines (%d unique) in %d files with text.\n" %
        [totalLines, uniqueLines, filesWithText]
end

  def UpdateScript(scriptFile, textFile)
    linesUpdated = 0

      File.open(textFile, 'rb') { |texts|
        File.open(scriptFile, 'r+b') { |script|
          mes = Mes.LoadFromStream script
          linesUpdated = mes.UpdateUsing texts

          script.rewind
          mes.SaveToStream script
        }
      }

    linesUpdated
  end

def UpdateScripts(files, sjsDir)
    totalFiles = 0
    linesUpdated = 0

  files.each { |file|
    print "Updating %s...\n" % File.basename(file)

      original = file + '~'
      textFile = sjsDir + '/' + File.basename(file, '.mes') + '.sjs'

    if !File.file? textFile
      print "  Skipped, no text file %s.\n" % File.basename(textFile)
      next
    elsif !File.file?(original) and File.mtime(file) + 60 >= File.mtime(textFile)
      print "  Skipped, judging from its modification time it's yet untranslated.\n"
      next
    end

      if File.file? original
        FileUtils.cp original, file
      else
        FileUtils.cp file, original
      end

    totalFiles += 1

    linesUpdated += UpdateScript file, textFile
  }

  print "\nDone, updated %d lines in %d files.\n" % [linesUpdated, totalFiles]
end


# --- Entry point ---

sjsDir = ARGV.first.to_s

if sjsDir.empty?
  print "\nMES text extractor/compiler by Proger_XP @ https://proger.me/vn\n"
    print "ruby MES.rb - extract texts\n"
    print "ruby MES.rb dir\\with\\sjs - update texts\n"
    print "In any case, *.mes must be present in '.' (CWD)\n"
    print "Tested with Ruby 1.9.3-p551 (2014-11-13) [i386-mingw32]\n"
    print "https://rubyinstaller.org/downloads/archives/\n"
    print "\n"

  StripAndSaveFromFiles( Dir['./*.mes'] )
else
  UpdateScripts( Dir['./*.mes'], sjsDir )
end