unit TrueWorldDLL;

interface

uses Windows, SysUtils, CRC32, DllShare, UnicodeUtils, StrConv, ExeUtils, FileStreamW;

type
  EVerifyGameExe = class (Exception)
  end;
    EChangeGameFonts = class (Exception)
    end;

function VerifyGameExe(RaiseIfInvalid: Boolean = True): Boolean;
function SJISCharToWide(uChar: DWord): WideChar;
procedure PatchSizeOfDrawnText(const Settings: TGameTranslationDllInitData);

var
  SpaceWidth, EllipsisWidth: DWord;
  ConvertJpnSymbols, NarrowKerning: Boolean;

implementation

function VerifyGameExe(RaiseIfInvalid: Boolean = True): Boolean;
  procedure CheckFailed(Message: String; FmtArgs: array of const);
  begin
    if RaiseIfInvalid then
      raise EVerifyGameExe.CreateFmt(Message, FmtArgs)
      else
        Result := False;
  end;

var
  Exe: TFileStreamW;
begin
  Exe := TFileStreamW.Create(ProcessFileName, fmOpenRead or fmShareDenyWrite);
  try
    if Exe.Size <> TargetExeSize then
      CheckFailed('Program file (%s) size must be %d bytes but it''s %d bytes.',
                 [BaseName(ProcessFileName), TargetExeSize, Exe.Size])
      else if CRC32OfStream(Exe) <> TargetExeCRC32 then
        CheckFailed('Checksum of program file (%s) is different (%.x8) from what it should have bern (%.8x).',
                    [BaseName(ProcessFileName), CRC32OfStream(Exe), TargetExeCRC32]);
  finally
    Exe.Free;
  end;
end;

function SJISCharToWide(uChar: DWord): WideChar;
var
  Wide: WideString;
begin
  if Hi(uChar) = 0 then     // $00xx in Shift-JIS is the same as in Windows-1250.
    case Lo(uChar) of
    $2D:    Wide := #$2014;   // em-dash.
    else
      Wide := Chr(uChar)
    end
    else if Hi(uChar) = $81 then
      if Lo(uChar) = $40 then  // $8140 = Japanese full-width space.
        Wide := ' '
        else if ConvertJpnSymbols then
          case Lo(uChar) of
          // $8163 = Shift-JIS ellipsis. It's converted to wide #$2026 but
          // on screen it looks strange because usually there are multiple
          // consequent ellipsises and they look strange due to surrounding
          // whitespace. But even with single dots spacing is too large
          // so currently MES.rb is replacing such lines beforoe compilation. 
          //$63:      Wide := '.';
//            $75, $76,
//            $77, $78: Wide := '"';
          $75:      Wide := Quotes[0];  // Japanese regular quote
            $76:    Wide := Quotes[1];
          $77:      Wide := Quotes[2];  // Japanese double/thick quote
            $78:    Wide := Quotes[3];
          $69:      Wide := '(';
            $6A:    Wide := ')';
          $5B, $5C: Wide := #$2014;     // em-dash.
          $48:      Wide := '?';
          $49:      Wide := '!';
          end;

  if Wide = '' then
    Wide := ToWideString(CP_SHIFTJIS, String(Chr(Hi(uChar)) + Chr(Lo(uChar))));
                                   
  if Wide = '' then
    Wide := ' ';      // protection against unexpected cases.
    
  Result := Wide[1];
//OutputDebugStringw(pwidechar(wideformat('SJISCharToWide: %04x -> %04x %s', [uchar, word(result), result])));
end;

var
  SleptStrings: array[0..$FFFF] of Boolean;

// From StringsW.pas.
function HashOf(Key: PChar): DWord;
begin
  Result := 0;
  while Key[0] <> #0 do
  begin
    Result := ((Result shl 2) or (Result shr (SizeOf(Result) * 8 - 2))) xor Ord(Key[0]);
    Inc(Key);
  end;
end;

function CharWidth(const SJISString: PChar; DC: HDC; FontSize: Integer; Index: Integer): DWord; register;
var
  Size: TSize;
  Ch: WideChar;
  Hash: Word;
begin
  Result := 0;
  Ch := SJISCharToWide(Byte(SJISString[Index + 1]) + (Byte(SJISString[Index + 0]) shl 8));
  if GetTextExtentPoint32W(DC, @Ch, 1, Size) then
  begin
    Result := Size.cx;
//    OutputDebugString(pchar(format(' cx = %d', [result])));
    if NarrowKerning then
      Result := Round(Result / 2);
  end;
  if (FontSize = $28) and PauseOnLargeFont then
  begin
    // We're using @l to not just change the font size but also to force a pause
    // between the letters. However, TextSizePatchProc is called not only during
    // initial rendering (when letters first appear) but also whenever screen
    // needs to be redrawn and when browsing the history. I couldn't find a way
    // to check at which phase we are so we're just remembering lines that we
    // output here and only sleeping for unique lines.
    Hash := HashOf(SJISString) and $FFFF;
    if not SleptStrings[Hash] then
    begin
      Sleep(100);
      // We mark the line (SJISString) as "output" after sleeping of its last
      // char - or when another '@' appears (e.g. '@n' at the end of every line
      // as inserted by MES.rb). Also by convention we assume @l is only effective
      // for 1 line (hence #10) since it'd take over half a second to output a
      // whole @l line - already too long for the reader.
      if (Word(SJISString + Index + 2) = 0) or (SJISString[Index + 2] = '@') or (SJISString[Index + 2] = #10) then
        SleptStrings[Hash] := True;
    end;
  end;
end;

procedure TextSizePatchProc;
const
  DcAddress = $48C480;
  GameFontSizeAddr = $92DFCC;
  TSizeLength = SizeOf(TSize);
  // See sub_4042C0.
  HFont1BAddr = $908DA0;
  HFont1CAddr = $908DAC;
  HFont28Addr = $908DB0;  
  HFont14Addr = $908DA8;

  JpnSpace    = $4081;
  JpnEllipsis = $6381;

  ReturnTo: Pointer = Pointer($417590);
asm
  PUSH EAX
  PUSH EBX
  PUSH EDX

    CMP WORD PTR DS:[EBX+ESI], JpnSpace

    JNE @NotASpace
      MOV ECX, SpaceWidth
      JMP @End

    @NotASpace:
      CMP WORD PTR DS:[EBX+ESI], JpnEllipsis

      JNE @NormalChar
        MOV ECX, EllipsisWidth
        JMP @End

    @NormalChar:
      // See sub_4042C0 for font selection.
      // 4 possible sizes: s/n/m/l (small/norm/med/large; 14 1B 1C 28).
      MOV ECX, DWORD PTR DS:[GameFontSizeAddr]
      CMP ECX, $1B
      JNZ @Font1
        MOV ECX, DWORD PTR DS:[HFont1BAddr]
        JMP @FontEnd
      @Font1:
      CMP ECX, $1C
      JNZ @Font2
        MOV ECX, DWORD PTR DS:[HFont1CAddr]    
        JMP @FontEnd
      @Font2:
      CMP ECX, $28
      JNZ @Font3
        MOV ECX, DWORD PTR DS:[HFont28Addr]
        JMP @FontEnd
      @Font3:
        MOV ECX, DWORD PTR DS:[HFont14Addr]
      @FontEnd:
      PUSH ECX

        MOV EDX, DWORD PTR DS:[DcAddress]
      PUSH EDX
      CALL SelectObject

//      LEA EAX, DWORD PTR DS:[EBX+ESI]   // EBX = game's array of char for line to output;
                                        // ESI - current char index.
      MOV EAX, EBX
      MOV EDX, DWORD PTR DS:[DcAddress]
      MOV ECX, DWORD PTR DS:[GameFontSizeAddr]
      PUSH ESI
      CALL CharWidth               // EAX = PChar
      MOV ECX, EAX
{
      SUB ESP, TSizeLength
        MOV EBX, ESP

        PUSH EBX                        // TSize ptr
        PUSH 1                          // str length
        PUSH EAX                        // str ptr           
        MOV ECX, DWORD PTR DS:[DcAddress]
        PUSH ECX                        // DC
        CALL GetTextExtentPoint32W

        CMP EAX, 0
        JZ @EndExtent
          MOV ECX, DWORD PTR DS:[EBX]     // EBX = TSize (X, Y: DWord)
//MOV DWORD PTR DS:[EBX], 18
//MOV DWORD PTR DS:[EBX+4], 18
//MOV ECX, 30
          CMP NarrowKerning, 0
          JZ @EndExtent
            SHR ECX, 1            // make spaces between chars more narrow.
        @EndExtent:
    ADD ESP, TSizeLength
}
  @End:

  POP EDX
  POP EBX
  POP EAX

  JMP ReturnTo
end;

procedure PatchSizeOfDrawnText(const Settings: TGameTranslationDllInitData);
const
  JumpFrom: Pointer = Pointer($41758A);
var
  Process: THandle;
  PatchProc: procedure;
  JMP: TJmpInstruction;
begin
  PatchProc := TextSizePatchProc;
    SpaceWidth := Settings.SpaceWidth;
    EllipsisWidth := Settings.EllipsisWidth;
    NarrowKerning := Settings.NarrowKerning;
    ConvertJpnSymbols := Settings.ConvertJpnSymbols;

  Process := OpenSelfProcess;
  try
    JMP := MakeJMP(JumpFrom, @PatchProc);
    WriteProcessMemory(Process, JumpFrom, JMP, SizeOf(JMP), 'Cannot patch size of text (JMP addr)');
  finally
    CloseHandle(Process);
  end;
end;

end.
