program xpcm2wav;

{$APPTYPE CONSOLE}

uses
  MMSystem,
  Classes,
  Windows,
  SysUtils,
  ShellAPI,
  UnicodeUtils,
  FileStreamW;

var
  flaccvt: widestring;

// info for the .wav file format got from https://ccrma.stanford.edu/courses/422/projects/WaveFormat/
procedure SaveTo(const tofile: string; const bgm: tstream; const offset, size: dword);
var
  outfile: tfilestream;
  riffchunk: packed record
    signature:  array[0..3] of char;
    chunksize:  dword;  // filesize - 8
    format:     array[0..3] of char;
  end;
  fmtsubchunk: packed record
    id:         array[0..3] of char;
    size:       dword;
    audiofmt:   word;
    channels:   word;
    samplerate: dword;
    byterate:   dword;
    blockalign: word;
    bitspersample: word;
  end;
  datasubchunk: packed record
    id:         array[0..3] of char;
    size:       dword;
  end;
begin
  bgm.Position := offset;

  outfile := TFileStream.Create(tofile, fmcreate);
  try
    outfile.Size := sizeof(riffchunk);
    outfile.Position := sizeof(riffchunk);

    with fmtsubchunk do
    begin
      id := 'fmt ';
      size := sizeof(fmtsubchunk) - 8;
      audiofmt := WAVE_FORMAT_PCM;
      channels := 2;
      samplerate := 44100;
      byterate := 176400;
      blockalign := 4;
      bitspersample := 16;
    end;
    outfile.Write(fmtsubchunk, sizeof(fmtsubchunk));

    datasubchunk.id := 'data';
    datasubchunk.size := size;
    outfile.Write(datasubchunk, sizeof(datasubchunk));

    outfile.CopyFrom(bgm, size);

    outfile.Position := 0;
      riffchunk.signature := 'RIFF';
      riffchunk.chunksize := outfile.Size - 8;
      riffchunk.format := 'WAVE';
    outfile.Write(riffchunk, sizeof(riffchunk));
  finally
    outfile.Free;
  end;
end;

function ExecAndWait(CmdLine: WideString; WinState: Word): DWord;
const
  Error = DWord(-1);
var
  StartInfo: TStartupInfo;
  ProcInfo: TProcessInformation;
begin
  Result := Error;

  FillChar(StartInfo, SizeOf(StartInfo), #0);
  with StartInfo do
  begin
    cb := SizeOf(StartInfo);
    dwFlags := STARTF_USESHOWWINDOW;
    wShowWindow := WinState
  end;

  if CreateProcessW(NIL, PWideChar(CmdLine), NIL, NIL, False, CREATE_NEW_CONSOLE or NORMAL_PRIORITY_CLASS,
                    NIL, PWideChar(ExtractFilePath(ParamStrW(0))), StartInfo, ProcInfo) then
  begin
    WaitForSingleObject(ProcInfo.hProcess, INFINITE);
      if not GetExitCodeProcess(ProcInfo.hProcess, Result) then
        Result := Error;

    CloseHandle(ProcInfo.hProcess);
    CloseHandle(ProcInfo.hThread);
  end;
end;

procedure PcmToWAV(SrcPCM, DestWAV: WideString);
  procedure SkipHeader(PCM: TStream; out SoundDataSize: DWord; out PcmType: Byte);
  var
    Sign: array[0..3] of Char;
  begin
    PCM.ReadBuffer(Sign, Length(Sign));
    if Sign <> 'XPCM' then
      raise Exception.Create('Wrong file signature - nto an XPCM sound?');

    PCM.ReadBuffer(SoundDataSize, 4);
    PCM.ReadBuffer(PcmType, 1);

      if not (PcmType in [1, 2, 3]) then
        PcmType := 0;  // any other value than 1-3 is considered as default by the game engine.

    PCM.Seek(3, soFromCurrent);

    case PcmType of
    0:    PCM.Seek($10, soFromCurrent);
    {1, 3:
      // is commented out because the sound data seems to be somehow prcoessed before waving out.
      begin
        PCM.Seek($10, soFromCurrent);
        PCM.ReadBuffer(SoundDataSize, 4);
      end}
    else
      raise Exception.CreateFmt('Sorry, this file is of different type (%d), only type 0 is supported now.',
                                [PcmType]);
    end;
  end;

var
  pcm: tfilestreamw;
  datasize: dword;
  pcmtype: byte;
begin
  writeln(basename(srcpcm), '...');

    destwav := expandfilename(destwav);

  pcm := tfilestreamw.Create(srcpcm, fmopenread or fmsharedenywrite);
  try
    try
      skipheader(pcm, datasize, pcmtype);
    except
      on E: Exception do
      begin
        writeln('  couldn''t convert this file: ', e.Message);
        exit;
      end;
    end;

    saveto(destwav, pcm, pcm.Position, datasize);

    if execandwait(wideformat('%s -f "%s"', [flaccvt, destwav]), sw_hide) = 0 then
    begin
      writeln('  ok, also converted to FLAC, deleting old WAV...');
      deletefilew(pwidechar(destwav));
    end
      else
        deletefilew(pwidechar( changefileext(destwav, '.flac') ));
  finally
    pcm.Free;
  end;
end;

procedure AllPcmToWavInDir(Src, Dest: WideString);
var
  SR: TWin32FindDataW;
  Handle: DWord;
  PcmFile: WideString;
begin
    Src := IncludeTrailingPathDelimiter(Src) + '*.pcm';
  Handle := FindFirstFileW(PWideChar(Src), SR);
  if Handle <> INVALID_HANDLE_VALUE then
  begin
    repeat
      PcmFile := WideString(SR.cFileName);
      if (ExtractFileExt(PcmFile) = '.pcm') then
        PcmToWAV( PcmFile, ChangeFileExt(PcmFile, '.wav') );
    until not FindNextFileW(Handle, SR);
    Windows.FindClose(Handle)
  end
end;

{ Main }

var
  fmtfile: string   = 'thbgm_95.fmt';
  wavefile: string  = 'thbgm.dat';
  outdir: string    = '.';

begin
  writeln('~ (X)PCM to WAV convertor ~'#9'by Proger_XP'#9'https://proger.me/vn');
  writeln(#9'30 July 2010');
  writeln;

  chdir( extractfilepath(paramstrw(0)) );
  flaccvt := 'flac.exe';

  try
    AllPcmToWavInDir('.', '.');
  except
    on E: Exception do
      writeln('Exception <', E.ClassName, '>:', #10, E.Message);
  end;

  writeln;
  writeln('Finished convertion of all *.pcm files in the program''s directory.');
  readln;
end.
